/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2011 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#include "generic_animations.h"
#include "dbus.h"
#include "anidata_p.h"

#include <kwinconfig.h>
#include <kconfiggroup.h>
#include <kwindowsystem.h>

#include <QDBusConnection>
#include <QTimerEvent>

#include <QtDebug>


using namespace KWin;
KWIN_EFFECT( generic_animations, GenericAnimations )

static Atom slideAtom;

#include "animation.defaults"

GenericAnimations::GenericAnimations() : AnimationEffect()
, m_lastActive(0)
, m_moveResizeWindow(0)
, m_resizeConnected(false)
{
    initDefaultAnimations();

    slideAtom = XInternAtom( display(), "_KDE_SLIDE", False );
    effects->registerPropertyType( slideAtom, true );
    // TODO hackish way to announce support, make better after 4.0
    unsigned char dummy = 0;
    XChangeProperty(display(), rootWindow(), slideAtom, slideAtom, 8, PropModeReplace, &dummy, 1);

    QDBusConnection::sessionBus().registerObject("/GenericAnimations", this);
    new GenericAnimationsAdaptor(this);

    connect( effects, SIGNAL(windowClosed(EffectWindow*)), SLOT(windowClosed(EffectWindow*)) );

    reconfigure( ReconfigureAll );
}

GenericAnimations::~GenericAnimations()
{
    effects->registerPropertyType( slideAtom, false );
}

void GenericAnimations::reconfigure( ReconfigureFlags )
{
    KConfigGroup conf = effects->effectConfig( "GenericAnimations" );

    effects->disconnect( SIGNAL(windowActivated(EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowAdded(EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowMinimized(EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowUnminimized(EffectWindow*)), this );
    effects->disconnect( SIGNAL(desktopChanged(int,int)), this );
    effects->disconnect( SIGNAL(desktopChanged(int,int,EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowOpacityChanged(EffectWindow*,qreal, qreal)), this );
    effects->disconnect( SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)), this );
//     effects->disconnect( SIGNAL(propertyNotify(EffectWindow*,long)), this );
    m_resizeConnected = false;

    m_animations.clear();

    QList<AniData> list;
    QStringList keys = conf.keyList();
    foreach (const QString &key, defaultAnimations.keys())
        if (!keys.contains(key))
            keys << key;
    foreach (const QString &key, keys) {
        if (key.endsWith("Animations")) {
            const QString str = conf.readEntry(key, defaultAnimations.value(key, QString()));
            list = AniData::list(str);
            if (!list.isEmpty())
                m_animations.insert(key.left(key.size() - 10), list);
        }
    }

    if ((m_animateMaximize = conf.readEntry("UnMaximize", true)))
    {
        connect( effects, SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)),
                           SLOT(windowGeometryShapeChanged( EffectWindow*, const QRect&)) );
        m_resizeConnected = true;
        connect( effects, SIGNAL(windowStartUserMovedResized(EffectWindow*)), SLOT(moveResizeStarted(EffectWindow*)) );
        connect( effects, SIGNAL(windowFinishUserMovedResized(EffectWindow*)), SLOT(moveResizeStopped(EffectWindow*)) );
    }
    if (conf.readEntry("Opacity", true))
        connect( effects, SIGNAL(windowOpacityChanged(EffectWindow*,qreal, qreal)), SLOT(opacityChanged(EffectWindow*,qreal, qreal)) );
    m_slidePopups = conf.readEntry("SlidingPopups", true);
//     if (m_slidePopups)
//         connect(effects, SIGNAL(propertyNotify(EffectWindow*,long)), this, SLOT(propertyChanged(EffectWindow*,long)));

    if (m_animations.contains("Activation") || m_animations.contains("ImplicitActivation") || m_animations.contains("Deactivation"))
        connect( effects, SIGNAL(windowActivated(EffectWindow*)), SLOT(windowActivated(EffectWindow*)) );
    if (m_slidePopups || m_animations.contains("Add"))
        connect( effects, SIGNAL(windowAdded(EffectWindow*)), SLOT(windowAdded(EffectWindow*)) );
    // "Close" is _always_ connected in constructor to hook on _before_ the baseclass!
    if (m_animations.contains("Minimize"))
        connect( effects, SIGNAL(windowMinimized(EffectWindow*)), SLOT(windowMinimized(EffectWindow*)) );
    if (m_animations.contains("Unminimize"))
        connect( effects, SIGNAL(windowUnminimized(EffectWindow*)), SLOT(windowUnminimized(EffectWindow*)) );

    if (m_animations.contains("LeavingDesktop") || m_animations.contains("EnteringDesktop") ||
        m_animations.contains("LeavingDesktopReverse") || m_animations.contains("EnteringDesktopReverse"))
    {
        if ( !connect( effects, SIGNAL(desktopChanged(int,int, EffectWindow*)), SLOT(desktopChanged(int,int, EffectWindow*)) ) )
            connect( effects, SIGNAL(desktopChanged(int,int)), SLOT(desktopChanged(int,int)) );

        if (!m_animateMaximize) // otherwise is already connected - TODO: do we need a flag to setup connections?
        {   // to avoid actions while dragging windows across desktops
            connect( effects, SIGNAL(windowStartUserMovedResized(EffectWindow*)), SLOT(moveResizeStarted(EffectWindow*)) );
            connect( effects, SIGNAL(windowFinishUserMovedResized(EffectWindow*)), SLOT(moveResizeStopped(EffectWindow*)) );
        }
    }

}


static bool matches( EffectWindow *w, NET::WindowTypeMask type)
{
    if ( (type & NET::NormalMask) && w->isNormalWindow() )
        return true;
    if ( (type & NET::PopupMenuMask) && w->isPopupMenu() )
        return true;
    if ( (type & NET::DialogMask) && w->isDialog() )
        return true;
    if ( (type & NET::MenuMask) && w->isMenu() )
        return true;
    if ( (type & NET::TooltipMask) && w->isTooltip() )
        return true;
    if ( (type & NET::DropdownMenuMask) && w->isDropdownMenu() )
        return true;
    if ( (type & NET::ComboBoxMask) && w->isComboBox() )
        return true;
    if ( (type & NET::DNDIconMask) && w->isDNDIcon() )
        return true;
    if ( (type & NET::UtilityMask) && w->isUtility() )
        return true;
    if ( (type & NET::SplashMask) && w->isSplash() )
        return true;
    if ( (type & NET::DockMask) && w->isDock() )
        return true;
    if ( (type & NET::ToolbarMask) && w->isToolbar() )
        return true;
    if ( (type & NET::NotificationMask) && w->isNotification() )
        return true;
    if ( (type & NET::DesktopMask) && w->isDesktop() )
        return true;
    if ( (type & NET::NotificationMask) && w->isNotification() )
        return true;

    return false;
// NET::Unknown
// NET::Override
// NET::TopMenu

// isManaged()
// isModal()
// isSpecialWindow()

}

static int mouseToAnchor(EffectWindow *w, int m, AnimationEffect::Attribute a)
{
    if ( !(m & AnimationEffect::Mouse) || (a != AnimationEffect::Scale && a != AnimationEffect::Position && a != AnimationEffect::Size) )
        return m;

    QRect geo = w->geometry();
    QPoint cursor = QCursor::pos();
    m = 0;
    if ( cursor.x() < geo.x() + geo.width()/4 )
        m |= AnimationEffect::Left;
    else if ( cursor.x() > geo.x() + 3*geo.width()/4 )
        m |= AnimationEffect::Right;
    if ( cursor.y() < geo.y() + geo.height()/4 )
        m |= AnimationEffect::Top;
    else if ( cursor.y() > geo.y() + 3*geo.height()/4 )
        m |= AnimationEffect::Bottom;

    return m;
}

void GenericAnimations::animationEnded(EffectWindow *w, Attribute a)
{
    if (a != Position)
        return;
    int i = m_slidingPopups.indexOf( w );
    if (i < 0)
        return;
    m_slidingPopups.removeAt(i);
    if (m_slidingPopups.isEmpty()) {
        EffectWindowList windows = effects->stackingOrder();
        foreach (EffectWindow *dock, windows) {
            if (dock->isDock())
                effects->setElevatedWindow(dock, false);
        }
    }
}

void GenericAnimations::opacityChanged( EffectWindow *w, qreal oldOpacity, qreal newOpacity )
{
    float diff = qAbs(newOpacity - oldOpacity);
    if (diff > 0.05)
        animate( w, Opacity, 0, qMax(0.0f,1-(1-diff)*(1-diff))*500, 1.0, QEasingCurve::Linear, 0, oldOpacity/newOpacity );
}

// void GenericAnimations::propertyChanged(EffectWindow *w, long atom)
// {
//     if (w && (Atom)atom == slideAtom)
//         slideInOutPopup(w, !w->isDeleted());
// }

#define IF_HAVE_ANIMATION(_ANI_) AniHash::const_iterator animation = m_animations.find(_ANI_); if (animation != m_animations.constEnd())
#define META mouseToAnchor(w, d.meta, d.attribute)

void GenericAnimations::trigger(QString key, WId id)
{
    IF_HAVE_ANIMATION(key) {
        EffectWindowList windows;
        if (id) {
            if (EffectWindow *w = effects->findWindow(id))
                windows << w;
            else
                return; // there's no such window
        }
        else
            windows = effects->stackingOrder();
        foreach (EffectWindow *w, windows) {
            foreach (const AniData &d, *animation)
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
        }
    }
}


void GenericAnimations::timerEvent(QTimerEvent *te)
{
    if (te->timerId() == m_implicitActivationTimer)
    {
        killTimer(m_implicitActivationTimer);
        m_implicitActivationTimer = 0;
        m_implicitActivationCandidate = 0;
    }
    else
        AnimationEffect::timerEvent(te);
}

static EffectWindow *slidingWindow = 0;
void GenericAnimations::desktopChanged(int oldDesktop, int newDesktop, EffectWindow *with)
{
    slidingWindow = with;
    desktopChanged(oldDesktop, newDesktop);
    slidingWindow = 0;
}

void GenericAnimations::desktopChanged(int oldDesktop, int newDesktop)
{
    if (m_moveResizeWindow)
        return;
    EffectWindowList allWindows = effects->stackingOrder();
    AniHash::const_iterator animation, leaving, entering;
    int d1 = newDesktop - oldDesktop, d2 = oldDesktop - newDesktop;
    if (d1 < 0) d1 += effects->numberOfDesktops();
    if (d2 < 0) d2 += effects->numberOfDesktops();
    if (d1 > d2) {
        leaving = m_animations.find("LeavingDesktopReverse");
        entering = m_animations.find("EnteringDesktopReverse");
    } else {
        leaving = m_animations.find("LeavingDesktop");
        entering = m_animations.find("EnteringDesktop");
    }
    foreach (EffectWindow *w, allWindows) {
        if (w->isMinimized() || w == slidingWindow)
            continue;
        const bool onOld = w->isOnDesktop(oldDesktop), onNew = w->isOnDesktop(newDesktop);
        if (onOld && !onNew)
            animation = leaving;
        else if (onNew && !onOld)
            animation = entering;
        else
            continue;
        if (animation != m_animations.constEnd()) {
            foreach (const AniData &d, *animation)
                if ( matches( w, d.windowType) )
                    animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
        }
    }
}

void GenericAnimations::moveResizeStarted(EffectWindow *w)
{
    m_moveResizeWindow = w;
    if (m_animateMaximize) {
        effects->disconnect( SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)), this );
        m_resizeConnected = false;
    }
}

void GenericAnimations::moveResizeStopped(EffectWindow *w)
{
    Q_UNUSED(w);
    m_moveResizeWindow = 0;
    if (m_animateMaximize && !m_resizeConnected) {
        connect( effects, SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)),
                          SLOT(windowGeometryShapeChanged( EffectWindow*, const QRect&)) );
        m_resizeConnected = true;
    }
}

static const int size_of_long = sizeof(long);
bool GenericAnimations::slideInOutPopup(EffectWindow *w, bool in)
{
    if (!m_slidePopups)
        return false;

    QByteArray slideData = w->readProperty( slideAtom, slideAtom, 32 );
    if (slideData.size() < 2*size_of_long)
        return false;

    const long *ld = reinterpret_cast<const long*>(slideData.constData());
    // "start = ld[ 0 ];"

    uint a;
    FPx2 other;
    QRect slideRect(w->geometry());
    switch (ld[1]) {
    case 0: // slide from/to left
        a = Right|Top;
        other.set(0, w->geometry().y());
        slideRect.setLeft(0);
        break;
    case 1:
        a = Bottom|Left;
        other.set(w->geometry().x(), 0);
        slideRect.setTop(0);
        break; // slide from/to top
    case 2:
        a = Left|Top;
        slideRect.setRight(effects->clientArea(ScreenArea , w).right());
        other.set(slideRect.right(), w->geometry().y());
        break; // slide from/to right
    case 3:
        a = Top|Left;
        slideRect.setBottom(effects->clientArea(ScreenArea , w).bottom());
        other.set(w->geometry().x(), slideRect.bottom());
        break; // slide from/to bottom
    default:
        return false;
    }

    uint meta(0);
    setMetaData(in ? SourceAnchor : TargetAnchor, a, meta);

    long time = 300;
    if (!in && slideData.size() >= 4*size_of_long)
        time = ld[3];
    else if (slideData.size() >= 3*size_of_long)
        time = ld[2];

    EffectWindowList windows = effects->stackingOrder();
    foreach (EffectWindow *dock, windows) {
        if (dock->isDock() && dock != w && !m_slidingPopups.contains(dock) && dock->geometry().intersects(slideRect))
            effects->setElevatedWindow(dock, true); // TODO: also set property if it's elevated otherwise?
    }

    // yes - append blindly! "if (!m_slidingPopups.contains(w))" is wrong because we can trigger one animation while the other is running
    // the first exiting would than redo elevations though in fact this window still needs one.
    m_slidingPopups << w;

    if (in)
        animate( w, Position, meta, time, FPx2(), QEasingCurve::OutQuart, 0, other );
    else
        animate( w, Position, meta, time, other, QEasingCurve::InQuart );

    return true;
}

void GenericAnimations::windowActivated( EffectWindow *w )
{
    if (m_lastActive && !(m_lastActive->isMinimized() || m_lastActive->isDeleted())) {
        IF_HAVE_ANIMATION("Deactivation") {
            foreach (const AniData &d, *animation)
                if ( matches( m_lastActive, d.windowType) )
                    animate( m_lastActive, d.attribute, mouseToAnchor(m_lastActive, d.meta, d.attribute), d.duration, d.to, d.curve, d.time, d.from );
        }
    }

    m_lastActive = w;

    if (w) {
        QString reason("Activation");
        if (w == m_implicitActivationCandidate)
        {
            m_implicitActivationCandidate = 0;
            reason = "ImplicitActivation";
        }
        IF_HAVE_ANIMATION(reason) {
            foreach (const AniData &d, *animation)
                if ( matches( w, d.windowType) )
                    animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
        }
    }
}

#define RESTART(_T_, _TT_) if (_T_) killTimer(_T_); _T_ = startTimer(_TT_)

void GenericAnimations::windowAdded( EffectWindow* w )
{
    if ( w->isMinimized() || w->isDeleted() || KWindowSystem::showingDesktop())
        return;

    if (!slideInOutPopup(w, true)) {
        IF_HAVE_ANIMATION("Add") {
            foreach (const AniData &d, *animation)
                if ( matches( w, d.windowType) )
                    animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
        }
    }
    m_implicitActivationCandidate = w;
    RESTART(m_implicitActivationTimer, 150);
}

void GenericAnimations::windowClosed( EffectWindow* w )
{
    if (slideInOutPopup(w, false))
        return;
    IF_HAVE_ANIMATION("Close") {
        foreach (const AniData &d, *animation)
            if ( matches( w, d.windowType) )
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
    }
}

void GenericAnimations::windowDeleted( EffectWindow* w )
{
    if (w == m_lastActive)
        m_lastActive = 0;
}


void GenericAnimations::windowGeometryShapeChanged( EffectWindow *w, const QRect &old )
{
    if (!m_animateMaximize || m_moveResizeWindow || w == m_implicitActivationCandidate)
        return;
    if (w->isDeleted() || w->isMinimized())
        return;
    if (!old.isValid())
        return;

    const int d = qAbs(w->width()-old.width()) + qAbs(w->height()-old.height());
    if (d < 24)
        return;

    const QRectF geo(w->geometry());
    bool growing = false;
    QEasingCurve::Type shape = QEasingCurve::Linear;
    float fx = geo.width() / old.width();
    if (fx > 1.0)
        { fx = 1.0/fx; growing = true; }

    float fy = geo.height() / old.height();
    if (fy > 1.0)
        { fy = 1.0/fy; growing = true; }

    float time = qMin(fx, fy);
    if (time < 0.5 && !growing)
        shape = QEasingCurve::OutQuad;
    time = qMax(0.0f, 1.1f-time*time)*350;

    animate( w, Position, 0, time, geo.center(), shape, 0, old.center() );
    animate( w, Size, 0, time, geo.size(), shape, 0, old.size() );
}

void GenericAnimations::windowMinimized( EffectWindow* w )
{
    IF_HAVE_ANIMATION("Minimize") {
        foreach (const AniData &d, *animation)
            if ( matches( w, d.windowType) )
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
    }
}

void GenericAnimations::windowUnminimized( EffectWindow* w )
{
    IF_HAVE_ANIMATION("Unminimize") {
        foreach (const AniData &d, *animation)
            if ( matches( w, d.windowType) )
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
    }
    m_implicitActivationCandidate = w;
    RESTART(m_implicitActivationTimer, 150);
}

#undef RESTART
#undef META

#include "moc_generic_animations.cpp"
#include "moc_dbus.cpp"
